from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn2pmml import _is_ordinal
from statsmodels.base.model import Model
from statsmodels.tools import add_constant

import numpy

class StatsModelsEstimator(BaseEstimator):

	def __init__(self, model_class, fit_intercept = True, **init_params):
		if not isinstance(model_class, type):
			raise TypeError("The model class object is not a Python class")
		if not issubclass(model_class, Model):
			raise TypeError("The model class is not a subclass of {}".format(Model.__name__))
		self.model_class = model_class
		self.fit_intercept = fit_intercept
		self.init_params = init_params

	def fit(self, X, y, **fit_params):
		if self.fit_intercept:
			X = add_constant(X, has_constant = "add")
		self.model_ = self.model_class(endog = y, exog = X, **self.init_params)
		fit_method = fit_params.pop("fit_method", "fit")
		if (fit_method != "fit") and (not fit_method.startswith("fit_")):
			raise ValueError()
		self.results_ = getattr(self.model_, fit_method)(**fit_params)
		return self

	def remove_data(self):
		self.results_.remove_data()

class StatsModelsClassifier(StatsModelsEstimator, ClassifierMixin):

	def __init__(self, model_class, fit_intercept = True, **init_params):
		super(StatsModelsClassifier, self).__init__(model_class = model_class, fit_intercept = fit_intercept, **init_params)

	def regression_table_(self):
		params = self.results_.params.T
		if len(params.shape) == 1:
			params = params.reshape((1, -1))
		return params

	@property
	def coef_(self):
		reg_table = self.regression_table_()
		if self.fit_intercept:
			return reg_table[:, 1:]
		return reg_table

	@property
	def intercept_(self):
		reg_table = self.regression_table_()
		if self.fit_intercept:
			return reg_table[:, 0]
		return numpy.zeros((reg_table.shape[0], ), dtype = float)

	def fit(self, X, y, **fit_params):
		classes, y_encoded = numpy.unique(y, return_inverse = True)
		self.classes_ = classes
		super(StatsModelsClassifier, self).fit(X = X, y = y_encoded, **fit_params)
		return self

	def predict(self, X, **predict_params):
		proba = self.predict_proba(X, **predict_params)
		indices = numpy.argmax(proba, axis = 1)
		return numpy.take(self.classes_, indices)

	def predict_proba(self, X, **predict_proba_params):
		if self.fit_intercept:
			X = add_constant(X, has_constant = "add")
		proba = self.results_.predict(X, **predict_proba_params)
		if proba.ndim == 1:
			proba = numpy.vstack((1 - proba, proba)).T
		return proba

class StatsModelsOrdinalClassifier(StatsModelsEstimator):

	def __init__(self, model_class, **init_params):
		super(StatsModelsOrdinalClassifier, self).__init__(model_class = model_class, fit_intercept = False, **init_params)

	def regression_table_(self):
		params = self.results_.params
		coef_params = params[0:-(len(self.classes_) - 1)]
		return coef_params.T

	@property
	def coef_(self):
		reg_table = self.regression_table_()
		return reg_table

	@property
	def intercept_(self):
		if hasattr(self.results_, "offset"):
			return self.results_.offset
		return 0.0

	@property
	def threshold_(self):
		params = self.results_.params
		th_params = params[-(len(self.classes_) - 1):]
		return th_params.T

	def fit(self, X, y, **fit_params):
		dtype = y.dtype
		if not _is_ordinal(dtype):
			raise TypeError()
		self.classes_ = numpy.asarray(dtype.categories)
		super(StatsModelsOrdinalClassifier, self).fit(X = X, y = y, **fit_params)
		return self

	def predict(self, X, **predict_params):
		proba = self.predict_proba(X, **predict_params)
		indices = numpy.argmax(proba, axis = 1)
		return numpy.take(self.classes_, indices)

	def predict_proba(self, X, **predict_proba_params):
		proba = self.results_.predict(X, **predict_proba_params)
		return proba

class StatsModelsRegressor(StatsModelsEstimator, RegressorMixin):

	def __init__(self, model_class, fit_intercept = True, **init_params):
		super(StatsModelsRegressor, self).__init__(model_class = model_class, fit_intercept = fit_intercept, **init_params)

	def regression_table_(self):
		return self.results_.params.T

	@property
	def coef_(self):
		reg_table = self.regression_table_()
		if self.fit_intercept:
			reg_table = reg_table[1:]
		return reg_table

	@property
	def intercept_(self):
		reg_table = self.regression_table_()
		if self.fit_intercept:
			return reg_table[0]
		return 0.0

	def predict(self, X, **predict_params):
		if self.fit_intercept:
			X = add_constant(X, has_constant = "add")
		return self.results_.predict(exog = X, **predict_params)



def add_options(estimator, **options):
	if hasattr(estimator, "pmml_options_"):
		pmml_options = dict(estimator.pmml_options_)
		pmml_options.update(options)
	else:
		pmml_options = dict(options)
	estimator.pmml_options_ = pmml_options

def clear_options(estimator):
	if hasattr(estimator, "pmml_options_"):
		delattr(estimator, "pmml_options_")



from lxml import etree
from pandas.api.types import is_string_dtype, is_integer_dtype

NS_PMML44 = "{http://www.dmg.org/PMML-4_4}"

def make_element(tag, **_extra):
	return etree.Element(NS_PMML44 + tag, **_extra)

class PMMLElement(object):

	def __init__(self, root):
		self.root = root

	def tostring(self, encoding = "utf-8"):
		return etree.tostring(self.root, pretty_print = True).decode(encoding)

	def set(self, name, value):
		self.root.set(name, str(value))
		return self

	def append(self, element):
		self.root.append(element.root if isinstance(element, PMMLElement) else element)
		return self

	def text(self, text):
		self.root.text = text

class Array(PMMLElement):

	def __init__(self, values):
		super(Array, self).__init__(make_element("Array"))

		if is_string_dtype(values):
			vtype = "string"
		elif is_integer_dtype(values):
			vtype = "int"
		else:
			vtype = "real"

		self.set("n", str(len(values)))
		self.set("type", vtype)
		self.text(" ".join([str(value) for value in values]))

class Extension(PMMLElement):

	def __init__(self, name = None, value = None):
		super(Extension, self).__init__(make_element("Extension"))

		if name is not None:
			self.set("name", name)
		if value is not None:
			self.set("value", value)



from pandas import CategoricalDtype
try:
	from sklearn_pandas import DataFrameMapper
except ImportError:
	class DataFrameMapper(object):
		pass
from sklearn.base import BaseEstimator
from sklearn.compose import ColumnTransformer
from sklearn.exceptions import NotFittedError
from sklearn.feature_selection import SelectFromModel, SelectorMixin
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn2pmml.resources import _package_classpath
from subprocess import PIPE, Popen
from zipfile import ZipFile

import dill
import joblib
import numpy
import os
import pandas
import platform
import re
import sklearn
import tempfile
import warnings

from .metadata import __copyright__, __license__, __version__
from .pipeline import PMMLPipeline

def _is_categorical(dtype):
	if dtype == object or dtype == str or dtype == bool:
		return True
	elif _is_pandas_categorical(dtype):
		return True
	return False

def _is_pandas_categorical(dtype):
	if hasattr(dtype, "name"):
		return dtype.name == "category"
	return False

def _is_proto_pandas_categorical(dtype):
	if isinstance(dtype, str) and dtype == "category":
		return True
	if isinstance(dtype, CategoricalDtype):
		return dtype.categories is None
	return False

def _is_ordinal(dtype):
	if _is_pandas_ordinal(dtype):
		return True
	return False

def _is_pandas_ordinal(dtype):
	if isinstance(dtype, CategoricalDtype):
		return dtype.ordered
	return False

class EstimatorProxy(BaseEstimator):

	def __init__(self, estimator, attr_names = ["feature_importances_"]):
		self.estimator = estimator
		self.attr_names = attr_names
		try:
			self._copy_attrs()
		except NotFittedError:
			pass

	def _copy_attrs(self):
		for attr_name in self.attr_names:
			if hasattr(self.estimator, attr_name):
				setattr(self, attr_name, getattr(self.estimator, attr_name))

	@property
	def classes_(self):
		return self.estimator.classes_

	def fit(self, X, y = None, **fit_params):
		self.estimator.fit(X, y, **fit_params)
		self._copy_attrs()
		return self

	def predict(self, X, **predict_params):
		return self.estimator.predict(X, **predict_params)

	def predict_proba(self, X, **predict_proba_params):
		return self.estimator.predict_proba(X, **predict_proba_params)

class SelectorProxy(BaseEstimator):

	def __init__(self, selector):
		self.selector = selector
		try:
			self._copy_attrs()
		except NotFittedError:
			pass

	def _copy_attrs(self):
		try:
			setattr(self, "support_mask_", self.selector._get_support_mask())
		except ValueError:
			pass

	def fit(self, X, y = None, **fit_params):
		self.selector.fit(X, y, **fit_params)
		self._copy_attrs()
		return self

	def transform(self, X):
		return self.selector.transform(X)

def _escape(obj, escape_func):
	if isinstance(obj, DataFrameMapper):
		obj.features = _escape_steps(obj.features, escape_func = escape_func)
		if hasattr(obj, "built_features"):
			if obj.built_features is not None:
				obj.built_features = _escape_steps(obj.built_features, escape_func = escape_func)
	elif isinstance(obj, ColumnTransformer):
		obj.transformers = _escape_steps(obj.transformers, escape_func = escape_func)
		obj.remainder = escape_func(obj.remainder, escape_func = escape_func)
		if hasattr(obj, "transformers_"):
			obj.transformers_ = _escape_steps(obj.transformers_, escape_func = escape_func)
	elif isinstance(obj, FeatureUnion):
		obj.transformer_list = _escape_steps(obj.transformer_list, escape_func = escape_func)
	elif isinstance(obj, Pipeline):
		obj.steps = _escape_steps(obj.steps, escape_func = escape_func)
	elif isinstance(obj, SelectorMixin):
		return SelectorProxy(obj)
	elif isinstance(obj, list):
		return [escape_func(e, escape_func = escape_func) for e in obj]
	return obj

def _escape_steps(steps, escape_func):
	return [(step[:1] + (escape_func(step[1], escape_func = escape_func), ) + step[2:]) for step in steps]

def make_pmml_pipeline(estimator, active_fields = None, target_fields = None, escape_func = _escape):
	"""Wraps a Scikit-Learn estimator or pipeline object into a PMML pipeline object.

	Parameters:
	----------
	estimator: BaseEstimator
		The estimator or pipeline object.

	active_fields: list of strings, optional
		Feature names. If missing, "x1", "x2", .., "xn" are assumed.

	target_fields: list of strings, optional
		Label name(s). If missing, "y" is assumed.

	"""
	if not isinstance(estimator, BaseEstimator):
		raise TypeError("The estimator object is not an instance of {0}".format(BaseEstimator.__name__))
	pipeline = PMMLPipeline([
		("estimator", estimator)
	])
	pipeline = _escape(pipeline, escape_func = escape_func)
	if active_fields is not None:
		pipeline.active_fields = numpy.asarray(active_fields)
	if target_fields is not None:
		pipeline.target_fields = numpy.asarray(target_fields)
	return pipeline

def _make_java_command(java_home, java_opts, java_args):
	result = []
	if java_home is not None:
		if not isinstance(java_home, str):
			raise ValueError()
		result.extend([os.path.join(java_home, "bin", "java")])
	else:
		result.extend(["java"])
	if java_opts is not None:
		if not isinstance(java_opts, list):
			raise ValueError()
		result.extend(java_opts)
	if not isinstance(java_args, list):
		raise ValueError()
	result.extend(java_args)
	return result

def _java_version(java_home = None):
	cmd = _make_java_command(java_home = java_home, java_opts = None, java_args = ["-version"])
	try:
		process = Popen(cmd, stdout = PIPE, stderr = PIPE, bufsize = 1, universal_newlines = True)
	except:
		return None
	output, error = process.communicate()
	retcode = process.poll()
	if retcode:
		return None
	return _parse_java_version(error)

def _parse_java_version(java_version):
	match = re.match(r"^(.*)\sversion\s\"(.*)\"(|\s.+)$", java_version, re.MULTILINE)
	if match:
		return (match.group(1), match.group(2))
	else:
		return None

def _classpath(user_classpath):
	return _package_classpath() + user_classpath

def _process_classpath(name, fun, user_classpath):
	jars = _classpath(user_classpath)
	for jar in jars:
		with ZipFile(jar, "r") as zipfile:
			try:
				zipentry = zipfile.getinfo(name)
			except KeyError:
				pass
			else:
				fun(zipfile.open(zipentry))

def _dill_dump(obj, prefix):
	fd, path = tempfile.mkstemp(prefix = (prefix + "-"), suffix = ".pkl")
	try:
		with open(path, "wb") as dill_file:
			dill.dump(obj, dill_file)
	finally:
		os.close(fd)
	return path

def _joblib_dump(obj, prefix):
	fd, path = tempfile.mkstemp(prefix = (prefix + "-"), suffix = ".pkl.z")
	try:
		joblib.dump(obj, path, compress = 3)
	finally:
		os.close(fd)
	return path

def _is_supported(estimator):
	clazz = estimator.__class__
	# LightGBM and XGBoost Booster objects
	if clazz.__module__.split(".")[0] in ["lightgbm", "xgboost"] and clazz.__name__ == "Booster":
		return True
	# H2O Estimator objects
	if hasattr(estimator, "download_mojo"):
		return True
	return isinstance(estimator, BaseEstimator)

def sklearn2pmml(estimator, pmml_path, with_repr = False, pmml_schema = None, java_home = None, java_opts = None, user_classpath = [], dump_flavour = "joblib", debug = False):
	"""Converts a fitted estimator or pipeline object to PMML.

	Parameters:
	----------
	estimator: BaseEstimator
		The estimator or pipeline object.

	pmml_path: string
		The file path where the PMML document should be saved.

	with_repr: boolean, optional
		If true, insert a string containing a printable representation of the estimator object into the PMML document.

	pmml_schema: string
		The PMML schema version for the PMML document.

	java_home: string, optional
		The path to Java installation directory.
		Functionally analogous to the JAVA_HOME environment variable.

	java_opts: list of strings, optional
		Java options.
		Functionally analogous to the JAVA_OPTS environment variable.

	user_classpath: list of strings, optional
		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.
		The SkLearn2PMML classpath is constructed by appending user JAR files to package JAR files.

	dump_flavour: string, optional
		The flavour of pickle dump files.

	debug: boolean, optional
		If true, print information about the conversion process.

	"""
	if debug:
		java_version = _java_version(java_home = java_home)
		if java_version is None:
			java_version = ("java", "N/A")
		print("python: {0}".format(platform.python_version()))
		print("sklearn2pmml: {0}".format(__version__))
		print("sklearn: {0}".format(sklearn.__version__))
		print("pandas: {0}".format(pandas.__version__))
		print("numpy: {0}".format(numpy.__version__))
		print("dill: {0}".format(dill.__version__))
		print("joblib: {0}".format(joblib.__version__))
		print("{0}: {1}".format(java_version[0], java_version[1]))
	if not _is_supported(estimator):
		raise TypeError("The estimator object is not an instance of {0}".format(BaseEstimator.__name__))
	# if isinstance(estimator, Pipeline):
	if hasattr(estimator, "_final_estimator"):
		final_estimator = estimator._final_estimator
	else:
		final_estimator = estimator
	dumps = []
	try:
		java_args = ["-cp", os.pathsep.join(_classpath(user_classpath)), "com.sklearn2pmml.Main"]
		if with_repr:
			estimator.repr_ = repr(estimator)
		# if isinstance(final_estimator, H2OEstimator):
		if hasattr(final_estimator, "download_mojo"):
			if dump_flavour != "dill":
				warnings.warn("Changing dump flavour to dill")
				dump_flavour = "dill"
			# Avoid MOJO (re-)download if the indicator attribute is set
			if not (hasattr(final_estimator, "_mojo_path") or hasattr(final_estimator, "_mojo_bytes")):
				mojo_path = final_estimator.download_mojo()
				dumps.append(mojo_path)
				final_estimator._mojo_path = mojo_path
		if dump_flavour == "dill":
			pkl_path = _dill_dump(estimator, "estimator")
		elif dump_flavour == "joblib":
			pkl_path = _joblib_dump(estimator, "estimator")
		else:
			raise ValueError("Dump flavour {0} not in {1}".format(dump_flavour, ["dill", "joblib"]))
		java_args.extend(["--pkl-input", pkl_path])
		dumps.append(pkl_path)
		java_args.extend(["--pmml-output", pmml_path])
		if pmml_schema:
			java_args.extend(["--pmml-schema", pmml_schema])
		cmd = _make_java_command(java_home = java_home, java_opts = java_opts, java_args = java_args)
		if debug:
			print("Executing command:\n{0}".format(" ".join(cmd)))
		try:
			process = Popen(cmd, stdout = PIPE, stderr = PIPE, bufsize = 1, universal_newlines = True)
		except OSError:
			raise RuntimeError("Java is not installed, or the Java executable is not on system path")
		output, error = process.communicate()
		retcode = process.poll()
		if debug or retcode:
			if len(output):
				print("Standard output:\n{0}".format(output))
			else:
				print("Standard output is empty")
			if len(error):
				print("Standard error:\n{0}".format(error))
			else:
				print("Standard error is empty")
		if retcode:
			raise RuntimeError("The SkLearn2PMML application has failed. The Java executable should have printed more information about the failure into its standard output and/or standard error streams")
	finally:
		if debug:
			print("Preserved dump file(s): {0}".format(" ".join(dumps)))
		else:
			for dump in dumps:
				os.remove(dump)

def _parse_properties(lines):
	splitter = re.compile(r"\s*=\s*")
	properties = dict()
	for line in lines:
		line = line.decode("UTF-8").rstrip()
		if line.startswith("#"):
			continue
		key, value = splitter.split(line)
		properties[key] = value
	return properties

def _format_properties(properties):
	return ["{0} = {1}\n".format(k, v) for k, v in properties.items()]

def _expand_complex_key(key):
	begin = key.find("(")
	end = key.find(")", begin + 1)
	if begin < 0 or end < 0:
		return [key]

	prefix = key[:begin]
	body = key[begin + 1:end]
	suffix = key[end + 1:]

	result = []
	parts = body.split("|")
	for part in parts:
		result += _expand_complex_key(prefix + part + suffix)
	return result

def _expand_mapping(mapping):
	result = dict()
	for k, v in mapping.items():
		pythonClazzes = _expand_complex_key(k)
		javaClazz = v
		for pythonClazz in pythonClazzes:
			result[pythonClazz] = (javaClazz if javaClazz else pythonClazz)
	return result

def load_class_mapping(user_classpath = []):
	"""Loads the class mapping.

	Parameters:
	----------
	user_classpath: list of strings, optional
		The paths to JAR files that provide custom Transformer, Selector and/or Estimator converter classes.

	Returns:
	-------
	mapping: dict
		Mapping from Python class names to Java converter class names.

	"""
	mapping = dict()
	func = lambda x: mapping.update(_expand_mapping(_parse_properties(x.readlines())))
	_process_classpath("META-INF/sklearn2pmml.properties", func, user_classpath)
	return mapping

def make_class_mapping_jar(mapping, path):
	"""Generates a class mapping JAR file.

	Parameters:
	----------
	mapping: dict of strings
		Mapping from Python class names to Java converter class names.

	path: string
		The path to output JAR file.

	"""
	lines = _format_properties(mapping)

	with ZipFile(path, mode = "w") as zipfile:
		zipfile.writestr("META-INF/sklearn2pmml.properties", "".join(lines))

def _strip_module(name):
	parts = name.split(".")
	if len(parts) > 1:
		parts.pop(-2)
		return ".".join(parts)
	return name



from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn2pmml.util import check_predicate, eval_rows, to_expr_func

class RuleSetClassifier(BaseEstimator, ClassifierMixin):

	def __init__(self, rules, default_score = None):
		for rule in rules:
			if type(rule) is not tuple:
				raise TypeError("Rule is not a tuple")
			if len(rule) != 2:
				raise TypeError("Rule is not a two-element (predicate, score) tuple")
			predicate, score = rule
			check_predicate(predicate)
		self.rules = rules
		self.default_score = default_score

	def fit(self, X, y = None):
		return self

	def predict(self, X):
		rules = [(to_expr_func(predicate), score) for (predicate, score) in self.rules]

		def _eval_row(x):
			for (expr_func, score) in rules:
				if expr_func(x):
					return score
			return self.default_score

		return eval_rows(X, _eval_row)



import pkg_resources

def _package_classpath():
	jars = []
	with open(pkg_resources.resource_filename("sklearn2pmml.resources", "classpath.txt")) as classpath:
		jars = [pkg_resources.resource_filename("sklearn2pmml.resources", jar_name.strip()) for jar_name in classpath.readlines()]
	return jars


from pandas import DataFrame
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import FeatureUnion
from sklearn2pmml.preprocessing import IdentityTransformer

import numpy

class Memory(object):

	def __init__(self, data = dict()):
		self.data = data

	def __getitem__(self, key):
		return self.data[key]

	def __setitem__(self, key, value):
		self.data[key] = value

	def __delitem__(self, key):
		del self.data[key]

	def __len__(self):
		if isinstance(self.data, DataFrame):
			return self.data.shape[1]
		return len(self.data)

	def __contains__(self, key):
		return key in self.data

	def __copy__(self):
		return self

	def __deepcopy__(self, memo):
		result = self
		memo[id(self)] = result
		return result

	def __getstate__(self):
		state = self.__dict__.copy()
		state["data"] = self.data.__class__()
		return state

	def __setstate__(self, state):
		self.__dict__.update(state)

	def clear(self):
		if isinstance(self.data, DataFrame):
			self.data.drop(columns = self.data.columns, inplace = True)
		else:
			self.data.clear()

def _set_position(transformers, position):
	if position == "first":
		return transformers
	elif position == "last":
		return transformers[1:] + transformers[0:1]
	else:
		raise ValueError()

def make_memorizer_union(memory, names, transform_only = True, position = "first"):
	transformers = [
		("memorizer", Memorizer(memory, names, transform_only = transform_only)),
		("identity", IdentityTransformer()),
	]
	transformers = _set_position(transformers, position = position)
	return FeatureUnion(transformers)

def make_recaller_union(memory, names, position = "first"):
	transformers = [
		("recaller", Recaller(memory, names)),
		("identity", IdentityTransformer())
	]
	transformers = _set_position(transformers, position = position)
	return FeatureUnion(transformers)

class _BaseMemoryManager(BaseEstimator, TransformerMixin):

	def __init__(self, memory, names):
		self.memory = memory
		if not isinstance(names, list):
			raise TypeError()
		self.names = names

class Memorizer(_BaseMemoryManager):

	def __init__(self, memory, names, transform_only = True):
		super(Memorizer, self).__init__(memory, names)
		self.transform_only = transform_only

	def memorize(self, X):
		if X.shape[1] != len(self.names):
			raise ValueError()
		for idx, name in enumerate(self.names):
			if isinstance(X, DataFrame):
				x = X.iloc[:, idx]
			else:
				x = X[:, idx]
			self.memory[name] = x.copy()
		return numpy.empty(shape = (X.shape[0], 0), dtype = int)

	def get_feature_names_out(self, input_features = None):
		return numpy.asarray([])

	def fit(self, X, y = None):
		if not self.transform_only:
			self.memorize(X)
		return self

	def transform(self, X):
		return self.memorize(X)

class Recaller(_BaseMemoryManager):

	def __init__(self, memory, names, clear_after = False):
		super(Recaller, self).__init__(memory, names)
		self.clear_after = clear_after

	def recall(self, X):
		result = []
		for idx, name in enumerate(self.names):
			x = self.memory[name]
			result.append(x.copy())
			if self.clear_after:
				del self.memory[name]
		return numpy.asarray(result).T

	def get_feature_names_out(self, input_features = None):
		return numpy.asarray(self.names)

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return self.recall(X)



from collections import defaultdict
try:
	from collections.abc import Hashable
except ImportError:
	from collections import Hashable
from datetime import datetime
from io import StringIO
from pandas import CategoricalDtype, DataFrame, Series
from scipy.interpolate import BSpline
from scipy.sparse import lil_matrix
from sklearn.base import BaseEstimator, TransformerMixin
try:
	# SkLearn 1.2.0+
	from sklearn.base import OneToOneFeatureMixin
except ImportError:
	from sklearn.base import _OneToOneFeatureMixin as OneToOneFeatureMixin
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn2pmml import _is_pandas_categorical, _is_proto_pandas_categorical
from sklearn2pmml.util import cast, check_expression, check_predicate, dt_transform, ensure_def, eval_rows, is_1d, to_1d, to_expr_func, to_numpy, Reshaper

import numpy
import pandas
import types
import warnings

def _regex_engine(pattern):
	try:
		import pcre
		return pcre.compile(pattern)
	except ImportError:
		warnings.warn("Perl Compatible Regular Expressions (PCRE) library is not available, falling back to built-in Regular Expressions (RE) library. Transformation results might not be reproducible between Python and PMML environments when using more complex patterns", Warning)
		import re
		return re.compile(pattern)

def _unique(X):
	nonmissing_mask = pandas.notnull(X)
	return numpy.unique(X[nonmissing_mask])

class Aggregator(BaseEstimator, TransformerMixin):
	"""Aggregate continuous data."""

	def __init__(self, function):
		functions = ["min", "max", "sum", "prod", "product", "mean", "avg"]
		if function not in functions:
			raise ValueError("Function {0} not in {1}".format(function, functions))
		self.function = function

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		if self.function == "min":
			Xt = numpy.nanmin(X, axis = 1) 
		elif self.function == "max":
			Xt = numpy.nanmax(X, axis = 1)
		elif self.function == "sum":
			Xt = numpy.nansum(X, axis = 1)
		elif self.function == "prod" or self.function == "product":
			Xt = numpy.nanprod(X, axis = 1)
		elif self.function == "mean" or self.function == "avg":
			Xt = numpy.nanmean(X, axis = 1)
		else:
			raise ValueError(self.function)
		return Xt.reshape((-1, 1))

class BSplineTransformer(BaseEstimator, TransformerMixin):

	def __init__(self, bspline):
		if not isinstance(bspline, BSpline):
			raise TypeError("The spline object is not an instance of {0}".format(BSpline.__name__))
		self.bspline = bspline

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		Xt = self.bspline(X1d)
		return Xt.reshape(X.shape)

class CastTransformer(BaseEstimator, TransformerMixin, OneToOneFeatureMixin):
	"""Change data type."""

	def __init__(self, dtype):
		if isinstance(dtype, str) and dtype.startswith("datetime64"):
			dtypes = ["datetime64[D]", "datetime64[s]"]
			if dtype not in dtypes:
				raise ValueError("Temporal data type {0} not in {1}".format(dtype, dtypes))
		self.dtype = dtype

	def fit(self, X, y = None):
		if _is_proto_pandas_categorical(self.dtype):
			X = to_numpy(X)
			self.dtype_ = CategoricalDtype(categories = _unique(X), ordered = False)
		else:
			self.dtype_ = self.dtype
		return self

	def transform(self, X):
		return cast(X, self.dtype_)

class CutTransformer(BaseEstimator, TransformerMixin):
	"""Bin continuous data to categorical."""

	def __init__(self, bins, right = True, labels = None, include_lowest = True):
		self.bins = bins
		self.right = right
		self.labels = labels
		self.include_lowest = include_lowest

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		Xt = pandas.cut(X1d, bins = self.bins, right = self.right, labels = self.labels, include_lowest = self.include_lowest)
		if _is_pandas_categorical(Xt.dtype):
			Xt = to_numpy(Xt)
		return Xt.reshape(X.shape)

class DataFrameConstructor(BaseEstimator, TransformerMixin):

	def __init__(self, columns, dtype):
		self.columns = columns
		self.dtype = dtype

	def get_feature_names_out(self, input_features = None):
		return numpy.asarray(self.columns)

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return DataFrame(X, columns = self.columns, dtype = self.dtype)

class SeriesConstructor(BaseEstimator, TransformerMixin, OneToOneFeatureMixin):

	def __init__(self, name, dtype):
		self.name = name
		self.dtype = dtype

	def get_feature_names_out(self, input_features = None):
		if self.name is not None:
			return numpy.asarray([self.name])
		return super(SeriesConstructor, self).get_feature_names_out(input_features = input_features)

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		return Series(X1d, name = self.name, dtype = self.dtype)

def _int(X):
	if numpy.isscalar(X):
		return int(X)
	else:
		return cast(X, int)

class DurationTransformer(BaseEstimator, TransformerMixin):
	"""Calculate time difference."""

	def __init__(self, year):
		if year < 1900:
			raise ValueError("Year {0} is earlier than 1900".format(year))
		self.year = year
		self.epoch = datetime(year, 1, 1, tzinfo = None)

	def _to_duration(self, td):
		raise NotImplementedError()

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		def to_int_duration(X):
			duration = self._to_duration(pandas.to_timedelta(X - self.epoch))
			return _int(duration)

		return dt_transform(X, to_int_duration)

class DaysSinceYearTransformer(DurationTransformer):
	"""Calculate the number of days since the epoch."""

	def __init__(self, year):
		super(DaysSinceYearTransformer, self).__init__(year)

	def _to_duration(self, td):
		return td.days

class SecondsSinceYearTransformer(DurationTransformer):
	"""Calculate the number of seconds since the epoch."""

	def __init__(self, year):
		super(SecondsSinceYearTransformer, self).__init__(year)

	def _to_duration(self, td):
		return td.total_seconds()

class SecondsSinceMidnightTransformer(BaseEstimator, TransformerMixin):
	"""Calculate the number of seconds since midnight."""

	def __init__(self):
		pass

	def _to_duration(self, td):
		return td.seconds

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		def to_int_duration(X):
			dt = pandas.to_datetime(X)
			duration = self._to_duration(dt - dt.normalize())
			return _int(duration)

		return dt_transform(X, to_int_duration)

class ExpressionTransformer(BaseEstimator, TransformerMixin):
	"""Transform data using a Python expression.

	Parameters:
	----------
	map_missing_to: scalar, optional
		The return value when any of the expression arguments is missing.

	default_value: scalar, optional
		The return value when the expression result is missing.

	invalid_value_treatment: string
		The action to take when the evaluation of the expression raises an error.
	"""

	def __init__(self, expr, map_missing_to = None, default_value = None, invalid_value_treatment = None, dtype = None):
		self.expr = check_expression(expr)
		self.map_missing_to = map_missing_to
		self.default_value = default_value
		invalid_value_treatments = ["return_invalid", "as_missing"]
		if (invalid_value_treatment is not None) and (invalid_value_treatment not in invalid_value_treatments):
			raise ValueError("Invalid value treatment {0} not in {1}".format(invalid_value_treatment, invalid_value_treatments))
		self.invalid_value_treatment = invalid_value_treatment
		self.dtype = dtype

	def _eval(self, X):
		expr_func = to_expr_func(self.expr)

		def _eval_row(x):
			# x is array-like (row vector)
			if (self.map_missing_to is not None) and ((pandas.isnull(x)).any()):
				return self.map_missing_to
			try:
				xt = expr_func(x)
			except ArithmeticError as ae:
				if self.invalid_value_treatment == "return_invalid":
					raise ae
				elif self.invalid_value_treatment == "as_missing":
					xt = None
				else:
					pass
			# xt is scalar
			if (self.default_value is not None) and (pandas.isnull(xt)):
				return self.default_value
			return xt

		# Evaluate in PMML compatibility mode
		with numpy.errstate(divide = "raise"):
			Xt = eval_rows(X, _eval_row, to_numpy = (not is_1d(X)), shape = (-1, 1))
		return Xt

	def fit(self, X, y = None):
		if _is_proto_pandas_categorical(self.dtype):
			Xt = self._eval(X)
			Xt = cast(Xt, self.dtype)
			dtype = Xt.dtype
		else:
			dtype = self.dtype
		self.dtype_ = dtype
		return self

	def transform(self, X):
		Xt = self._eval(X)
		if hasattr(self, "dtype_"):
			dtype = self.dtype_
		else:
			if _is_proto_pandas_categorical(self.dtype):
				raise NotFittedError()
			dtype = self.dtype
		if dtype is not None:
			Xt = cast(Xt, dtype)
		return Xt

	def fit_transform(self, X, y = None):
		Xt = self._eval(X)
		if self.dtype is not None:
			Xt = cast(Xt, self.dtype)
			if _is_proto_pandas_categorical(self.dtype):
				dtype = Xt.dtype
			else:
				dtype = self.dtype
		else:
			dtype = None
		self.dtype_ = dtype
		return Xt

class IdentityTransformer(BaseEstimator, TransformerMixin, OneToOneFeatureMixin):
	"""Passes data through as-is."""

	def __init__(self):
		pass

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return X

class DateTimeFormatter(BaseEstimator, TransformerMixin):
	"""Formats dates, times and datetimes according to a pattern. Analogous to C's strftime() function.

	Parameters:
	----------
	pattern: string
		A POSIX-compliant formatting pattern.
	"""

	def __init__(self, pattern):
		self.pattern = pattern

	def _strftime(self, x):
		return x.strftime(self.pattern)

	def fit(self, X):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		func = lambda x: self._strftime(x)
		Xt = eval_rows(X1d, func, shape = X.shape)
		return Xt

class NumberFormatter(BaseEstimator, TransformerMixin):
	"""Formats numbers according to a pattern. Analogous to C's printf() function.

	Parameters:
	----------
	pattern: string
		A POSIX-compliant formatting pattern.
	"""

	def __init__(self, pattern):
		self.pattern = pattern

	def _printf(self, x):
		with StringIO() as buffer:
			print(self.pattern % (x), sep = "", end = "", file = buffer)
			return buffer.getvalue()

	def fit(self, X):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		func = lambda x: self._printf(x)
		Xt = eval_rows(X1d, func, shape = X.shape)
		return Xt

class LookupTransformer(BaseEstimator, TransformerMixin):
	"""Re-map 1D categorical data.

	If the mapping is not found, returns `default_value`.

	See also:
	--------
	FilterLookupTransformer

	"""

	def __init__(self, mapping, default_value):
		if type(mapping) is not dict:
			raise TypeError("Input value to output value mapping is not a dict")
		k_type = None
		v_type = None
		for k, v in mapping.items():
			if pandas.isnull(k):
				raise ValueError("Key is a missing value")
			if k_type is None:
				k_type = type(k)
			else:
				if type(k) != k_type:
					raise TypeError("Key is not a {0}".format(k_type.__name__))
			if pandas.isnull(v):
				continue
			if v_type is None:
				v_type = type(v)
			else:
				if type(v) != v_type:
					raise TypeError("Value is not a {0}".format(v_type.__name__))
		self.mapping = mapping
		if default_value is not None:
			if v_type is not None:
				if type(default_value) != v_type:
					raise TypeError("Default value is not a {0}".format(v_type.__name__))
		self.default_value = default_value

	def _transform_dict(self):
		transform_dict = defaultdict(lambda: self.default_value)
		transform_dict.update(self.mapping)
		return transform_dict

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		transform_dict = self._transform_dict()

		def _eval_row(x):
			if pandas.isnull(x):
				return x
			return transform_dict[x]

		Xt = eval_rows(X1d, _eval_row, shape = X.shape)
		return Xt

class FilterLookupTransformer(LookupTransformer):
	"""Selectively re-map 1D categorical data.

	If the mapping is not found, returns the original value unchanged.

	See also:
	--------
	LookupTransformer

	"""

	def __init__(self, mapping):
		super(FilterLookupTransformer, self).__init__(mapping, default_value = None)
		for k, v in mapping.items():
			if pandas.isnull(v):
				raise ValueError("Value is a missing value")
			if type(k) != type(v):
				raise TypeError("Key and Value type mismatch")

	def _transform_dict(self):
		return self.mapping

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		X1d = to_1d(X)
		transform_dict = self._transform_dict()

		def _eval_row(x):
			if x not in transform_dict:
				return x
			return transform_dict[x]

		Xt = eval_rows(X1d, _eval_row, shape = X.shape)
		return Xt

def _deeptype(t):
	return tuple([type(e) for e in t])

class MultiLookupTransformer(LookupTransformer):
	"""Re-map multidimensional categorical data."""

	def __init__(self, mapping, default_value):
		super(MultiLookupTransformer, self).__init__(mapping, default_value)
		k_type = None
		for k, v in mapping.items():
			if not isinstance(k, tuple):
				raise TypeError("Key is not a tuple")
			if k_type is None:
				k_type = _deeptype(k)
			else:
				if _deeptype(k) != k_type:
					raise TypeError("Key is not a tuple of {0}", ", ".join(e.__name__ for e in k_type))

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		transform_dict = self._transform_dict()

		def _eval_row(x):
			if (pandas.isnull(x)).any():
				return None
			# See https://stackoverflow.com/a/3460747
			# See https://stackoverflow.com/a/3338368
			x = x if isinstance(x, Hashable) else tuple(numpy.squeeze(numpy.asarray(x)))
			return transform_dict[tuple(x)]

		Xt = eval_rows(X, _eval_row, to_numpy = True)
		return Xt.reshape((-1, 1))

def _make_index(values):
	result = {}
	for i, v in enumerate(list(values)):
		result[v] = i
	if len(result) != len(values):
		raise ValueError()
	return result

class PMMLLabelBinarizer(BaseEstimator, TransformerMixin):
	"""Binarize categorical data in a missing value-aware way."""

	def __init__(self, sparse_output = False):
		self.sparse_output = sparse_output

	def fit(self, X, y = None):
		X1d = to_1d(X)
		self.classes_ = _unique(X1d)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		mapping = _make_index(self.classes_)
		if self.sparse_output:
			Xt = lil_matrix((len(X1d), len(mapping)), dtype = int)
		else:
			Xt = numpy.zeros((len(X1d), len(mapping)), dtype = int)
		for i, v in enumerate(X1d):
			if (pandas.notnull(v)) and (v in mapping):
				Xt[i, mapping[v]] = 1
		if self.sparse_output:
			Xt = Xt.tocsr()
		return Xt

class PMMLLabelEncoder(BaseEstimator, TransformerMixin):
	"""Encode categorical data in a missing value-aware way."""

	def __init__(self, missing_values = None):
		self.missing_values = missing_values

	def fit(self, X, y = None):
		X1d = to_1d(X)
		self.classes_ = _unique(X1d)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		mapping = _make_index(self.classes_)
		Xt = numpy.array([self.missing_values if pandas.isnull(v) else mapping.get(v, self.missing_values) for v in X1d])
		return Xt.reshape((-1, 1))

class PowerFunctionTransformer(BaseEstimator, TransformerMixin):
	"""Raise numeric data to power."""

	def __init__(self, power):
		if not isinstance(power, int):
			raise TypeError("Power {0} is not an integer".format(power))
		self.power = power

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		return numpy.power(X, self.power)

class ConcatTransformer(BaseEstimator, TransformerMixin):
	"""Concat data to string."""

	def __init__(self, separator = ""):
		self.separator = separator

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		func = lambda x: self.separator.join([str(v) for v in x])
		Xt = eval_rows(X, func, to_numpy = True)
		return Xt.reshape((-1, 1))

class MatchesTransformer(BaseEstimator, TransformerMixin):
	"""Match RE pattern."""

	def __init__(self, pattern):
		self.pattern = pattern

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		engine = _regex_engine(self.pattern)
		func = lambda x: bool(engine.search(x))
		Xt = eval_rows(X1d, func, shape = X.shape)
		return Xt

class ReplaceTransformer(BaseEstimator, TransformerMixin):
	"""Replace all RE pattern matches."""

	def __init__(self, pattern, replacement):
		self.pattern = pattern
		self.replacement = replacement

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		engine = _regex_engine(self.pattern)
		func = lambda x: engine.sub(self.replacement, x)
		Xt = eval_rows(X1d, func, shape = X.shape)
		return Xt

class StringNormalizer(BaseEstimator, TransformerMixin):
	"""Normalize the case and surrounding whitespace."""

	def __init__(self, function = None, trim_blanks = True):
		functions = ["lower", "lowercase", "upper", "uppercase"]
		if (function is not None) and (function not in functions):
			raise ValueError("Function {0} not in {1}".format(function, functions))
		self.function = function
		self.trim_blanks = trim_blanks

	def fit(self, X, y = None):
		return self

	def transform(self, X):
		Xt = cast(X, "U")
		# Transform
		if self.function is None:
			pass
		elif self.function == "lower" or self.function == "lowercase":
			Xt = numpy.char.lower(Xt)
		elif self.function == "upper" or self.function == "uppercase":
			Xt = numpy.char.upper(Xt)
		else:
			raise ValueError(self.function)
		# Trim blanks
		if self.trim_blanks:
			Xt = numpy.char.strip(Xt)
		return Xt

class SubstringTransformer(BaseEstimator, TransformerMixin):
	"""Extract substring."""

	def __init__(self, begin, end):
		if begin < 0:
			raise ValueError("Begin position {0} is negative".format(begin))
		if end < begin:
			raise ValueError("End position {0} is smaller than begin position {1}".format(end, begin))
		self.begin = begin
		self.end = end

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		func = lambda x: x[self.begin:self.end]
		Xt = eval_rows(X1d, func, shape = X.shape)
		return Xt

class WordCountTransformer(BaseEstimator, TransformerMixin):
	"""Count tokens."""

	def __init__(self, word_pattern = "\w+", non_word_pattern = "\W+"):
		self.word_pattern = word_pattern
		self.non_word_pattern = non_word_pattern
		self.pipeline_ = Pipeline([
			("word_replacer", ReplaceTransformer(pattern = "({0})".format(word_pattern), replacement = "1")),
			("non_word_replacer", ReplaceTransformer(pattern = "({0})".format(non_word_pattern), replacement = "")),
			("counter", ExpressionTransformer("len(X[0])", dtype = int))
		])

	def fit(self, X, y = None):
		to_1d(X)
		return self

	def transform(self, X):
		X1d = to_1d(X)
		# The expression "X]0]" assumes a two-dimensional array
		X1d = to_numpy(X1d).reshape((-1, 1))
		return self.pipeline_.transform(X1d)

def _to_sparse(X, step_mask, step_result):
	# Make array
	result = numpy.empty((X.shape[0], step_result.shape[1]), dtype = object)
	# Fill array
	result[step_mask.ravel(), :] = step_result
	return result

class SelectFirstTransformer(BaseEstimator, TransformerMixin):

	def __init__(self, steps, controller = None, eval_rows = True):
		for step in steps:
			if type(step) is not tuple:
				raise TypeError("Step is not a tuple")
			if len(step) != 3:
				raise TypeError("Step is not a three-element (name, transformer, predicate) tuple")
			name, transformer, predicate = step
			check_predicate(predicate)
		self.steps = steps
		if controller:
			if not hasattr(controller, "transform"):
				raise TypeError()
		self.controller = controller
		self.eval_rows = eval_rows

	def _to_evaluation_dataset(self, X):
		if self.controller is not None:
			return self.controller.transform(X)
		return X

	def _eval_step_mask(self, X, predicate):
		step_func = to_expr_func(predicate)
		if self.eval_rows:
			return eval_rows(X, step_func, dtype = bool)
		else:
			return step_func(X)

	def fit(self, X, y = None):
		X_eval = self._to_evaluation_dataset(X)
		mask = numpy.full(X.shape[0], fill_value = False)
		for name, transformer, predicate in self.steps:
			step_mask = numpy.logical_not(mask)
			step_mask_eval = self._eval_step_mask(X_eval[step_mask], predicate)
			if numpy.sum(step_mask_eval) < 1:
				raise ValueError(predicate)
			step_mask[step_mask] = step_mask_eval
			step_X = X[step_mask]
			step_y = y[step_mask] if y is not None else None
			transformer.fit(step_X, step_y)
			mask = numpy.logical_or(mask, step_mask)
		return self

	def transform(self, X):
		result = None
		X_eval = self._to_evaluation_dataset(X)
		mask = numpy.full(X.shape[0], fill_value = False)
		for name, transformer, predicate in self.steps:
			step_mask = numpy.logical_not(mask)
			step_mask_eval = self._eval_step_mask(X_eval[step_mask], predicate)
			if numpy.sum(step_mask_eval) < 1:
				continue
			step_mask[step_mask] = step_mask_eval
			step_X = X[step_mask]
			step_result = transformer.transform(step_X)
			step_result = _to_sparse(X, step_mask, step_result)
			if result is None:
				result = step_result
			else:
				result[step_mask] = step_result[step_mask]
			mask = numpy.logical_or(mask, step_mask)
		return result
